##############################################################################
#
# Copyright (c) 2001 Zope Foundation and Contributors.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
""" Tests for the db module
"""
import unittest

from six.moves._thread import get_ident

from .base import DB_CONN_STRING
from .base import DB_PASSWORD
from .base import DB_USER
from .base import NO_MYSQL_MSG
from .base import TABLE_COL_INT
from .base import TABLE_COL_VARCHAR
from .base import TABLE_NAME
from .base import MySQLRequiredLayer
from .base import PatchedConnectionTestsBase
from .base import have_test_database
from .dummy import FakeConnection


class DbFunctionsTests(unittest.TestCase):

    def test_DateTime_or_None(self):
        from DateTime.DateTime import DateTime

        from Products.ZMySQLDA.db import DateTime_or_None

        self.assertIsInstance(DateTime_or_None(None), DateTime)
        self.assertIsInstance(DateTime_or_None(DateTime()), DateTime)
        self.assertIsInstance(DateTime_or_None('1970-01-01 00:00:00'),
                              DateTime)
        self.assertIsNone(DateTime_or_None(''))


class DBPoolTests(unittest.TestCase):

    def _makeOne(self, *args, **kw):
        from Products.ZMySQLDA.db import DB
        from Products.ZMySQLDA.db import DBPool
        return DBPool(DB, **kw)

    def test_instantiate_defaults(self):
        from Products.ZMySQLDA.db import DB
        pool = self._makeOne()

        self.assertIs(pool._db_cls, DB)
        self.assertEqual(pool._db_pool, {})
        self.assertFalse(pool._create_db)
        self.assertFalse(pool.use_unicode)
        self.assertIsNone(pool.charset)
        self.assertIsNone(pool.timeout)

    def test_instantiate_use_unicode(self):
        pool = self._makeOne(create_db=True, use_unicode=True)

        self.assertTrue(pool._create_db)
        self.assertTrue(pool.use_unicode)
        self.assertIsNone(pool.charset)

    def test_instantiate_with_charset(self):
        pool = self._makeOne(charset='latin1')

        self.assertFalse(pool.use_unicode)
        self.assertEqual(pool.charset, 'latin1')

    def test_closeConnection(self):
        pool = self._makeOne()

        pool._db_pool[get_ident()] = 'foo'
        self.assertEqual(pool._db_pool, {get_ident(): 'foo'})
        pool.closeConnection()
        self.assertEqual(pool._db_pool, {})

        # This should not raise an error
        pool.closeConnection()

    def test_close(self):
        pool = self._makeOne()

        pool._db_pool['foo'] = 1
        pool._db_pool['bar'] = 2
        self.assertDictEqual(pool._db_pool, {'foo': 1, 'bar': 2})
        pool.close()
        self.assertEqual(pool._db_pool, {})

    def test__pool_get_set_del(self):
        pool = self._makeOne()

        self.assertIsNone(pool._pool_get('foo'))
        pool._pool_set('foo', 1)
        self.assertEqual(pool._pool_get('foo'), 1)
        pool._pool_del('foo')
        self.assertIsNone(pool._pool_get('foo'))

    def test_name(self):
        pool = self._makeOne()
        pool._db_flags = {'kw_args': {'db': 'foo_database'}}
        self.assertEqual(pool.name(), 'foo_database')


class PatchedDBPoolTests(PatchedConnectionTestsBase):

    def _makeOne(self, *args, **kw):
        from Products.ZMySQLDA.db import DB
        from Products.ZMySQLDA.db import DBPool
        return DBPool(DB, **kw)

    def test_variables(self):
        pool = self._makeOne()
        pool._db_flags = {'kw_args': {}}
        self.assertEqual(pool.variables(),
                         {'var1': 'val1', 'version': '5.5.5'})


@unittest.skipUnless(have_test_database(), NO_MYSQL_MSG)
class RealConnectionDBPoolTests(unittest.TestCase):

    layer = MySQLRequiredLayer

    def tearDown(self):
        if getattr(self, 'dbpool', None) is not None:
            self.dbpool.close()

    def _makeOne(self, conn_string=DB_CONN_STRING, **kw):
        from Products.ZMySQLDA.db import DB
        from Products.ZMySQLDA.db import DBPool
        dbpool = DBPool(DB, **kw)
        # Connect to the database. Stupid API.
        return dbpool(conn_string)

    def test_connect_bad_db(self):
        try:
            from _mysql_exceptions import OperationalError
        except ImportError:  # mysqlclient > 1.4
            from MySQLdb import OperationalError

        # DB doesn't exist and should not be created
        conn_str = 'notexisting %s %s' % (DB_USER, DB_PASSWORD)
        self.assertRaises(OperationalError, self._makeOne, conn_str,
                          use_unicode=True, create_db=False)

        # DB doesn't exist and should be created
        conn_str = 'notexisting %s %s' % (DB_USER, DB_PASSWORD)
        self.assertRaises(OperationalError, self._makeOne, conn_str,
                          use_unicode=True, create_db=True)

    def test_tables(self):
        self.dbpool = self._makeOne()

        tables = self.dbpool.tables()
        test_table = [x for x in tables if x['table_name'] == TABLE_NAME][0]
        self.assertEqual(test_table['table_name'], TABLE_NAME)
        self.assertTrue(test_table.get('description'))  # Details not needed

    def test_columns(self):
        self.dbpool = self._makeOne()
        cols = self.dbpool.columns(TABLE_NAME)
        self.assertEqual(len(cols), 2)
        self.assertEqual(cols[0]['name'], TABLE_COL_INT)
        self.assertEqual(cols[1]['name'], TABLE_COL_VARCHAR)

    def test_variables(self):
        self.dbpool = self._makeOne()
        mysql_vars = self.dbpool.variables()
        self.assertIsInstance(mysql_vars, dict)
        self.assertGreater(len(mysql_vars), 0)

    def test_string_literal(self):
        self.dbpool = self._makeOne()
        self.assertEqual(self.dbpool.string_literal('foo'), b"'foo'")

    def test_unicode_literal(self):
        self.dbpool = self._makeOne()
        self.assertEqual(self.dbpool.unicode_literal(u'foo'), b"'foo'")

        self.dbpool = self._makeOne(charset='utf8mb4')
        self.assertEqual(self.dbpool.unicode_literal(u'foo'), b"'foo'")


class DBTests(PatchedConnectionTestsBase):

    def _makeOne(self, **kw):
        from Products.ZMySQLDA.db import DB
        return DB(**kw)

    def test_instantiation_defaults(self):
        db = self._makeOne(kw_args={})
        self.assertIsNone(db.connection)
        self.assertEqual(db._kw_args, {})
        self.assertIsNone(db._mysql_lock)
        self.assertIsNone(db._use_TM)
        self.assertIsNone(db._transactions)
        self.assertIsInstance(db.db, FakeConnection)

    def test_instantiation(self):
        db = self._makeOne(connection='conn', kw_args={'a': 1}, use_TM=True,
                           mysql_lock=True, transactions=True)
        self.assertEqual(db.connection, 'conn')
        self.assertEqual(db._kw_args, {'a': 1})
        self.assertTrue(db._mysql_lock)
        self.assertTrue(db._use_TM)
        self.assertTrue(db._transactions)

    def test_close(self):
        db = self._makeOne(kw_args={})
        db.close()
        self.assertIsNone(db.db)

    def test__parse_connection_string_empty(self):
        db = self._makeOne(kw_args={})

        parsed = db._parse_connection_string('')
        self.assertEqual(parsed['connection'], '')
        self.assertFalse(parsed['use_TM'])
        self.assertFalse('try_transactions' in parsed)
        self.assertFalse('mysql_lock' in parsed)
        self.assertFalse('db' in parsed['kw_args'])
        self.assertFalse('host' in parsed['kw_args'])
        self.assertFalse('port' in parsed['kw_args'])
        self.assertFalse('user' in parsed['kw_args'])
        self.assertFalse('passwd' in parsed['kw_args'])
        self.assertFalse('unix_socket' in parsed['kw_args'])
        self.assertFalse(parsed['kw_args']['use_unicode'])
        self.assertFalse('charset' in parsed['kw_args'])

    def test__parse_connection_string_simple(self):
        db = self._makeOne(kw_args={})

        c_str = 'foo_db foo_user foo_pw'
        parsed = db._parse_connection_string(c_str)
        self.assertEqual(parsed['connection'], c_str)
        self.assertFalse(parsed['use_TM'])
        self.assertFalse(parsed['try_transactions'])
        self.assertFalse(parsed['mysql_lock'])
        self.assertEqual(parsed['kw_args']['db'], 'foo_db')
        self.assertFalse('host' in parsed['kw_args'])
        self.assertFalse('port' in parsed['kw_args'])
        self.assertEqual(parsed['kw_args']['user'], 'foo_user')
        self.assertEqual(parsed['kw_args']['passwd'], 'foo_pw')
        self.assertFalse('unix_socket' in parsed['kw_args'])
        self.assertFalse(parsed['kw_args']['use_unicode'])
        self.assertFalse('charset' in parsed['kw_args'])

    def test__parse_connection_string_explicit_host(self):
        db = self._makeOne(kw_args={})

        c_str = 'foo_db@127.0.0.1:3306 foo_user foo_pw'
        parsed = db._parse_connection_string(c_str)
        self.assertEqual(parsed['connection'], c_str)
        self.assertFalse(parsed['use_TM'])
        self.assertFalse(parsed['try_transactions'])
        self.assertFalse(parsed['mysql_lock'])
        self.assertEqual(parsed['kw_args']['db'], 'foo_db')
        self.assertEqual(parsed['kw_args']['host'], '127.0.0.1')
        self.assertEqual(parsed['kw_args']['port'], 3306)
        self.assertEqual(parsed['kw_args']['user'], 'foo_user')
        self.assertEqual(parsed['kw_args']['passwd'], 'foo_pw')
        self.assertFalse('unix_socket' in parsed['kw_args'])
        self.assertFalse(parsed['kw_args']['use_unicode'])
        self.assertFalse('charset' in parsed['kw_args'])

    def test__parse_connection_string_unix_socket(self):
        db = self._makeOne(kw_args={})

        c_str = 'foo_db foo_user foo_pw /tmp/mysql.sock'
        parsed = db._parse_connection_string(c_str)
        self.assertEqual(parsed['connection'], c_str)
        self.assertFalse(parsed['use_TM'])
        self.assertFalse(parsed['try_transactions'])
        self.assertFalse(parsed['mysql_lock'])
        self.assertEqual(parsed['kw_args']['db'], 'foo_db')
        self.assertFalse('host' in parsed['kw_args'])
        self.assertFalse('port' in parsed['kw_args'])
        self.assertEqual(parsed['kw_args']['user'], 'foo_user')
        self.assertEqual(parsed['kw_args']['passwd'], 'foo_pw')
        self.assertEqual(parsed['kw_args']['unix_socket'], '/tmp/mysql.sock')
        self.assertFalse(parsed['kw_args']['use_unicode'])
        self.assertFalse('charset' in parsed['kw_args'])

    def test__parse_connection_string_mysql_lock(self):
        db = self._makeOne(kw_args={})

        c_str = '*mylock foo_db@127.0.0.1:3306 foo_user foo_pw'
        parsed = db._parse_connection_string(c_str)
        self.assertEqual(parsed['connection'], c_str)
        self.assertTrue(parsed['use_TM'])
        self.assertFalse(parsed['try_transactions'])
        self.assertEqual(parsed['mysql_lock'], 'mylock')
        self.assertEqual(parsed['kw_args']['db'], 'foo_db')
        self.assertEqual(parsed['kw_args']['host'], '127.0.0.1')
        self.assertEqual(parsed['kw_args']['port'], 3306)
        self.assertEqual(parsed['kw_args']['user'], 'foo_user')
        self.assertEqual(parsed['kw_args']['passwd'], 'foo_pw')
        self.assertFalse('unix_socket' in parsed['kw_args'])
        self.assertFalse(parsed['kw_args']['use_unicode'])
        self.assertFalse('charset' in parsed['kw_args'])

    def test__parse_connection_string_transactions(self):
        db = self._makeOne(kw_args={})

        c_str = '+foo_db@127.0.0.1:3306 foo_user foo_pw'
        parsed = db._parse_connection_string(c_str)
        self.assertEqual(parsed['connection'], c_str)
        self.assertFalse(parsed['use_TM'])
        self.assertEqual(parsed['try_transactions'], '+')
        self.assertFalse(parsed['mysql_lock'])
        self.assertEqual(parsed['kw_args']['db'], 'foo_db')
        self.assertEqual(parsed['kw_args']['host'], '127.0.0.1')
        self.assertEqual(parsed['kw_args']['port'], 3306)
        self.assertEqual(parsed['kw_args']['user'], 'foo_user')
        self.assertEqual(parsed['kw_args']['passwd'], 'foo_pw')
        self.assertFalse('unix_socket' in parsed['kw_args'])
        self.assertFalse(parsed['kw_args']['use_unicode'])
        self.assertFalse('charset' in parsed['kw_args'])

    def test__parse_connection_string_use_unicode(self):
        db = self._makeOne(kw_args={})

        c_str = '+foo_db@127.0.0.1:3306 foo_user foo_pw'
        parsed = db._parse_connection_string(c_str, use_unicode=True)
        self.assertTrue(parsed['kw_args']['use_unicode'])
        self.assertEqual(parsed['kw_args']['charset'], 'utf8')

    def test__parse_connection_string_use_unicode_charset(self):
        db = self._makeOne(kw_args={})

        c_str = '+foo_db@127.0.0.1:3306 foo_user foo_pw'
        parsed = db._parse_connection_string(c_str, use_unicode=True,
                                             charset='latin1')
        self.assertTrue(parsed['kw_args']['use_unicode'])
        self.assertEqual(parsed['kw_args']['charset'], 'latin1')

    def test__parse_connection_string_charset(self):
        db = self._makeOne(kw_args={})

        c_str = '+foo_db@127.0.0.1:3306 foo_user foo_pw'
        parsed = db._parse_connection_string(c_str, charset='utf8')
        self.assertFalse(parsed['kw_args']['use_unicode'])
        self.assertEqual(parsed['kw_args']['charset'], 'utf8')

    def test_variables(self):
        db = self._makeOne(kw_args={})
        self.assertDictEqual(db.variables(),
                             {'var1': 'val1', 'version': '5.5.5'})

    def test__register(self):
        db = self._makeOne(kw_args={})
        self.assertFalse(db._registered)
        self.assertFalse(db._finalize)
        self.assertFalse(getattr(db, '_transaction_begun', False))

        db._register()
        self.assertTrue(db._registered)
        self.assertFalse(db._finalize)
        self.assertTrue(getattr(db, '_transaction_begun', False))

    def test__begin(self):
        db = self._makeOne(kw_args={})
        db._begin()
        self.assertTrue(db._transaction_begun)
        self.assertIsNone(db.db.last_query)

    def test__begin_transactions(self):
        db = self._makeOne(kw_args={})
        db._transactions = True
        db._begin()
        self.assertTrue(db._transaction_begun)
        self.assertEqual(db.db.last_query, 'BEGIN')

    def test__begin_mysql_lock(self):
        db = self._makeOne(kw_args={})
        db._mysql_lock = 'foo_lock'
        db._begin()
        self.assertTrue(db._transaction_begun)
        self.assertEqual(db.db.last_query, "SELECT GET_LOCK('foo_lock',0)")

    def test__finish(self):
        db = self._makeOne(kw_args={})

        # Without a running transaction, _abort does nothing.
        db._transaction_begun = False
        self.assertIsNone(db._finish())

        db._begin()
        db._finish()
        self.assertFalse(db._transaction_begun)
        self.assertIsNone(db.db.last_query)

    def test__finish_transactions(self):
        db = self._makeOne(kw_args={})
        db._transactions = True
        db._begin()
        db._finish()
        self.assertFalse(db._transaction_begun)
        self.assertEqual(db.db.last_query, 'COMMIT')

    def test__finish_mysql_lock(self):
        db = self._makeOne(kw_args={})
        db._mysql_lock = 'foo_lock'
        db._begin()
        db._finish()
        self.assertFalse(db._transaction_begun)
        self.assertEqual(db.db.last_query, "SELECT RELEASE_LOCK('foo_lock')")

    def test__abort(self):
        db = self._makeOne(kw_args={})

        # Without a running transaction, _abort does nothing.
        db._transaction_begun = False
        self.assertIsNone(db._abort())

        db._begin()
        db._abort()
        self.assertFalse(db._transaction_begun)
        self.assertIsNone(db.db.last_query)

    def test__abort_transactions(self):
        db = self._makeOne(kw_args={})
        db._transactions = True
        db._begin()
        db._abort()
        self.assertFalse(db._transaction_begun)
        self.assertEqual(db.db.last_query, 'ROLLBACK')

    def test__abort_mysql_lock(self):
        db = self._makeOne(kw_args={})
        db._mysql_lock = 'foo_lock'
        db._begin()
        db._abort()
        self.assertFalse(db._transaction_begun)
        self.assertEqual(db.db.last_query, "SELECT RELEASE_LOCK('foo_lock')")

    def test_savepoint_outside_transaction(self):
        db = self._makeOne(kw_args={})

        # If a savepoint is used outside of a transaction: AttributeError
        self.assertRaises(AttributeError, db.savepoint)

    def test_savepoint(self):
        db = self._makeOne(kw_args={})

        db._begin()
        sp = db.savepoint()
        self.assertEqual(db.db.last_query, 'SAVEPOINT %s' % sp.ident)


@unittest.skipUnless(have_test_database(), NO_MYSQL_MSG)
class RealConnectionDBTests(unittest.TestCase):

    layer = MySQLRequiredLayer

    def tearDown(self):
        self.db.close()

    def _makeOne(self):
        from Products.ZMySQLDA.db import DB
        flags = DB._parse_connection_string(DB_CONN_STRING)
        if 'try_transactions' in flags:
            del flags['try_transactions']
        return DB(**flags)

    def test_tables(self):
        self.db = self._makeOne()

        tables = self.db.tables()
        test_table = [x for x in tables if x['table_name'] == TABLE_NAME][0]
        self.assertEqual(test_table['table_name'], TABLE_NAME)
        self.assertTrue(test_table.get('description'))  # Details not needed

    def test_columns(self):
        self.db = self._makeOne()

        cols = self.db.columns(TABLE_NAME)
        self.assertEqual(len(cols), 2)
        self.assertEqual(cols[0]['name'], TABLE_COL_INT)
        self.assertEqual(cols[0]['type'], 'int')
        self.assertEqual(cols[0]['scale'], 10)
        self.assertFalse(cols[0]['nullable'])
        self.assertTrue(cols[0]['index'])
        self.assertTrue(cols[0]['primary_key'])
        self.assertTrue(cols[0]['unique'])
        self.assertEqual(cols[0]['key'], 'PRI')
        self.assertEqual(cols[0]['default'], '0000000000')
        self.assertEqual(cols[1]['name'], TABLE_COL_VARCHAR)
        self.assertEqual(cols[1]['type'], 'varchar')
        self.assertEqual(cols[1]['scale'], 20)

    def test_columns_badtable(self):
        self.db = self._makeOne()

        # Asking for columns from a bad table should just return empty results.
        self.assertFalse(self.db.columns('notexistingtable'))

    def test_query_error(self):
        try:
            from _mysql_exceptions import ProgrammingError
        except ImportError:  # mysqlclient > 1.4
            from MySQLdb import ProgrammingError
        self.db = self._makeOne()

        self.assertRaises(ProgrammingError, self.db.query,
                          'SELECT * from NOTEXISTING')


class _SavePointTests(unittest.TestCase):

    def _makeOne(self):
        from Products.ZMySQLDA.db import _SavePoint
        return _SavePoint(FakeConnection())

    def test_initialization(self):
        sp = self._makeOne()
        self.assertIsInstance(sp.db_conn, FakeConnection)
        self.assertEqual(sp.db_conn.last_query, 'SAVEPOINT %s' % sp.ident)

    def test_rollback(self):
        sp = self._makeOne()
        sp.rollback()
        self.assertEqual(sp.db_conn.last_query, 'ROLLBACK TO %s' % sp.ident)


def test_suite():
    return unittest.TestSuite((unittest.makeSuite(DbFunctionsTests),
                               unittest.makeSuite(DBPoolTests),
                               unittest.makeSuite(PatchedDBPoolTests),
                               unittest.makeSuite(RealConnectionDBPoolTests),
                               unittest.makeSuite(DBTests),
                               unittest.makeSuite(RealConnectionDBTests),
                               unittest.makeSuite(_SavePointTests)))



##############################################################################
#
# Copyright (c) 2001 Zope Foundation and Contributors.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
""" Dummy fixtures for testing
"""

import six


RESULTS = {'show table status': [['table1', 'engine1', None, None, 5, None,
                                  None, None, None, None, None, None, None,
                                  None, 'my_collation']],
           'show variables': [('var1', 'val1'), ('version', '5.5.5')]}

TABLE = {'table_name': 'table1', 'table_type': 'type1', 'description': ''}

COLUMNS = [{'name': 'col1', 'icon': 'icon1', 'description': 'desc1'},
           {'name': 'col2', 'icon': 'icon2', 'description': 'desc2'}]


class FakeColumns(object):

    def __init__(self, table_name):
        self.cols = {table_name: COLUMNS}

    def columns(self, table_name):
        return self.cols[table_name]


class FakeResults:

    def __init__(self, results, **kw):
        self.results = results
        self.next_index = 0

    def fetch_row(self, count):
        return self.results


class FakeConnection:

    def __init__(self, **kw):
        self.server_capabilities = 0
        self.last_results = None
        self.last_query = None
        self.string_literal_called = False
        self.unicode_literal_called = False

        for k, v in kw.items():
            setattr(self, k, v)

    def ping(self, *args):
        pass

    def query(self, sql):
        self.last_query = sql
        sql = sql.lower()
        self.last_results = FakeResults(RESULTS.get(sql, []))
        return self.last_results

    _query = query

    def store_result(self):
        return self.last_results

    def close(self):
        pass

    def string_literal(self, txt):
        self.string_literal_called = txt
        return txt

    def unicode_literal(self, txt):
        self.unicode_literal_called = txt
        if isinstance(txt, six.text_type):
            txt = txt.encode('UTF-8')
        return txt



##############################################################################
#
# Copyright (c) 2001 Zope Foundation and Contributors.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
import logging
import time

import MySQLdb
import six
from MySQLdb.constants import CLIENT
from MySQLdb.constants import CR
from MySQLdb.constants import ER
from MySQLdb.constants import FIELD_TYPE
from MySQLdb.converters import conversions
from six.moves._thread import allocate_lock
from six.moves._thread import get_ident

import transaction
from DateTime.DateTime import DateTime
from DateTime.interfaces import DateTimeError
from Shared.DC.ZRDB.TM import TM
from ZODB.POSException import ConflictError
from ZODB.POSException import TransactionFailedError


try:
    import _mysql
    from _mysql_exceptions import NotSupportedError
    from _mysql_exceptions import OperationalError
    from _mysql_exceptions import ProgrammingError
except ImportError:  # mysqlclient > 1.4
    import MySQLdb as _mysql
    from MySQLdb import NotSupportedError
    from MySQLdb import OperationalError
    from MySQLdb import ProgrammingError


LOG = logging.getLogger('ZMySQLDA')

hosed_connection = {
    CR.SERVER_GONE_ERROR: 'Server gone.',
    CR.SERVER_LOST: 'Server lost.',
    CR.COMMANDS_OUT_OF_SYNC: 'Commands out of sync. Possibly a misplaced'
                             ' semicolon (;) in a query.',
}

query_syntax_error = (ER.BAD_FIELD_ERROR,)

key_types = {'PRI': 'PRIMARY KEY', 'MUL': 'INDEX', 'UNI': 'UNIQUE'}

field_icons = 'bin', 'date', 'datetime', 'float', 'int', 'text', 'time'

icon_xlate = {
    'varchar': 'text',
    'char': 'text',
    'enum': 'what',
    'set': 'what',
    'double': 'float',
    'numeric': 'float',
    'blob': 'bin',
    'mediumblob': 'bin',
    'longblob': 'bin',
    'tinytext': 'text',
    'mediumtext': 'text',
    'longtext': 'text',
    'timestamp': 'datetime',
    'decimal': 'float',
    'smallint': 'int',
    'mediumint': 'int',
    'bigint': 'int',
}

type_xlate = {
    'double': 'float',
    'numeric': 'float',
    'decimal': 'float',
    'smallint': 'int',
    'mediumint': 'int',
    'bigint': 'int',
    'int': 'int',
    'float': 'float',
    'timestamp': 'datetime',
    'datetime': 'datetime',
    'time': 'datetime',
}


def DateTime_or_None(s):
    try:
        return DateTime(s)
    except DateTimeError:
        return None


class DBPool(object):
    """
      This class is an interface to the database connection..
      Its caracteristic is that an instance of this class interfaces multiple
      instanes of db_cls class, each one being bound to a specific thread.
    """

    connected_timestamp = ''
    _create_db = False
    use_unicode = False
    charset = None
    timeout = None

    def __init__(self, db_cls, create_db=False, use_unicode=False,
                 charset=None, timeout=None):
        """ Set transaction managed class for use in pool.
        """
        self._db_cls = db_cls
        # pool of one db object/thread
        self._db_pool = {}
        self._db_lock = allocate_lock()
        # auto-create db if not present on server
        self._create_db = create_db
        # unicode settings
        self.use_unicode = use_unicode
        self.charset = charset
        # timeout setting
        self.timeout = int(timeout) if timeout else None

    def __call__(self, connection):
        """ Parse the connection string.

            Initiate a trial connection with the database to check
            transactionality once instead of once per db_cls instance.

            Create database if option is enabled and database doesn't exist.
        """
        self.connection = connection
        db_flags = self._db_cls._parse_connection_string(connection,
                                                         self.use_unicode,
                                                         charset=self.charset,
                                                         timeout=self.timeout)
        self._db_flags = db_flags

        # connect to server to determin tranasactional capabilities
        # can't use db_cls instance as it requires this information to work
        try:
            connection = MySQLdb.connect(**db_flags['kw_args'])
        except OperationalError:
            if self._create_db:
                kw_args = db_flags.get('kw_args', {}).copy()
                db = kw_args.pop('db', None)
                if not db:
                    raise
                connection = MySQLdb.connect(**kw_args)
                create_query = 'create database %s' % db
                if self.use_unicode and not self.charset:
                    create_query += ' default character set %s' % (
                                    self._db_cls.unicode_charset)
                elif self.charset:
                    create_query += ' default character set %s' % self.charset
                connection.query(create_query)
                connection.store_result()
            else:
                raise
        transactional = connection.server_capabilities & CLIENT.TRANSACTIONS
        connection.close()

        # Some tweaks to transaction/locking db_flags based on server setup
        if db_flags['try_transactions'] == '-':
            transactional = False
        elif not transactional and db_flags['try_transactions'] == '+':
            raise NotSupportedError('transactions not supported by the server')
        db_flags['transactions'] = transactional
        del db_flags['try_transactions']
        if transactional or db_flags['mysql_lock']:
            db_flags['use_TM'] = True

        # will not be 100% accurate in regard to per thread connections
        # but as close as we're going to get it.
        self.connected_timestamp = DateTime()

        # return self as the database connection object
        # (assigned to _v_database_connection)
        return self

    def closeConnection(self):
        """ Close this threads connection. Used when DA is being reused
            but the connection string has changed. Need to close the db_cls
            instances and recreate to the new database. Only have to worry
            about this thread as when each thread hits the new connection
            string in the DA this method will be called.
        """
        ident = get_ident()
        try:
            self._pool_del(ident)
        except KeyError:
            pass

    def close(self):
        """ Used when manually closing the database. Resetting the pool
            dereferences the db_cls instances where they are then collected
            and closed.
        """
        self._db_pool = {}

    def _pool_set(self, key, value):
        """ Add a db to pool.
        """
        self._db_lock.acquire()
        try:
            self._db_pool[key] = value
        finally:
            self._db_lock.release()

    def _pool_get(self, key):
        """ Get db from pool. Read only, no lock necessary.
        """
        return self._db_pool.get(key)

    def _pool_del(self, key):
        """ Remove db from pool.
        """
        self._db_lock.acquire()
        try:
            del self._db_pool[key]
        finally:
            self._db_lock.release()

    def name(self):
        """ Return name of database connected to.
        """
        return self._db_flags['kw_args']['db']

    # Passthrough aliases for methods on db_cls class.
    def variables(self, *args, **kw):
        return self._access_db(method_id='variables', args=args, kw=kw)

    def tables(self, *args, **kw):
        return self._access_db(method_id='tables', args=args, kw=kw)

    def columns(self, *args, **kw):
        return self._access_db(method_id='columns', args=args, kw=kw)

    def query(self, *args, **kw):
        return self._access_db(method_id='query', args=args, kw=kw)

    def string_literal(self, *args, **kw):
        return self._access_db(method_id='string_literal', args=args, kw=kw)

    def unicode_literal(self, *args, **kw):
        try:
            return self._access_db(method_id='unicode_literal',
                                   args=args, kw=kw)
        except AttributeError:  # mysqlclient > 1.3.11
            if self.charset and self.charset.startswith('utf8'):
                charset = 'UTF-8'
            else:
                charset = self.charset
            # This is modeled after code in MySQLdb.connections.__init__
            new_args = (args[0].encode(charset or 'latin1'),) + args[1:]
            return self._access_db(method_id='string_literal',
                                   args=new_args, kw=kw)

    def _access_db(self, method_id, args, kw):
        """
          Generic method to call pooled objects' methods.
          When the current thread had never issued any call, create a db_cls
          instance.
        """
        ident = get_ident()
        db = self._pool_get(ident)
        if db is None:
            db = self._db_cls(**self._db_flags)
            self._pool_set(ident, db)
        return getattr(db, method_id)(*args, **kw)


class DB(TM):

    Database_Error = _mysql.Error

    defs = {
        FIELD_TYPE.CHAR: 'i',
        FIELD_TYPE.DATE: 'd',
        FIELD_TYPE.DATETIME: 'd',
        FIELD_TYPE.DECIMAL: 'n',
        FIELD_TYPE.NEWDECIMAL: 'n',
        FIELD_TYPE.DOUBLE: 'n',
        FIELD_TYPE.FLOAT: 'n',
        FIELD_TYPE.INT24: 'i',
        FIELD_TYPE.LONG: 'i',
        FIELD_TYPE.LONGLONG: 'l',
        FIELD_TYPE.SHORT: 'i',
        FIELD_TYPE.TIMESTAMP: 'd',
        FIELD_TYPE.TINY: 'i',
        FIELD_TYPE.YEAR: 'i',
    }

    conv = conversions.copy()
    conv[FIELD_TYPE.LONG] = int
    conv[FIELD_TYPE.DATETIME] = DateTime_or_None
    conv[FIELD_TYPE.DATE] = DateTime_or_None
    conv[FIELD_TYPE.DECIMAL] = float
    conv[FIELD_TYPE.NEWDECIMAL] = float
    del conv[FIELD_TYPE.TIME]

    _p_oid = _p_changed = None
    _sort_key = '1'
    _registered = False
    _finalize = False

    unicode_charset = 'utf8'  # hardcoded for now

    def __init__(self, connection=None, kw_args=None, use_TM=None,
                 mysql_lock=None, transactions=None):
        self.connection = connection  # backwards compat
        self._kw_args = kw_args
        self._mysql_lock = mysql_lock
        self._use_TM = use_TM
        self._transactions = transactions
        self._forceReconnection()

    def close(self):
        """ Close connection and dereference.
        """
        if getattr(self, 'db', None):
            self.db.close()
            self.db = None

    __del__ = close

    def _forceReconnection(self):
        """ (Re)Connect to database.
        """
        try:  # try to clean up first
            self.db.close()
        except Exception:
            pass
        self.db = MySQLdb.connect(**self._kw_args)
        # Newer mysqldb requires ping argument to attmept a reconnect.
        # This setting is persistent, so only needed once per connection.
        self.db.ping(True)

    @classmethod
    def _parse_connection_string(cls, connection, use_unicode=False,
                                 charset=None, timeout=None):
        """ Done as a class method to both allow access to class attribute
            conv (conversion) settings while allowing for wrapping pool class
            use of this method. The former is important to allow for subclasses
            to override the conv settings while the latter is important so
            the connection string doesn't have to be parsed for each instance
            in the pool.
        """
        kw_args = {'conv': cls.conv}
        flags = {'kw_args': kw_args, 'connection': connection}
        kw_args['use_unicode'] = use_unicode
        if use_unicode:
            kw_args['charset'] = cls.unicode_charset
        if charset:
            kw_args['charset'] = charset
        items = connection.split()
        flags['use_TM'] = None
        if _mysql.get_client_info()[0] >= '5':
            kw_args['client_flag'] = CLIENT.MULTI_STATEMENTS
        if items:
            lockreq, items = items[0], items[1:]
            if lockreq[0] == '*':
                flags['mysql_lock'] = lockreq[1:]
                db_host, items = items[0], items[1:]
                flags['use_TM'] = True  # redundant. eliminate?
            else:
                flags['mysql_lock'] = None
                db_host = lockreq
            if '@' in db_host:
                db, host = db_host.split('@', 1)
                kw_args['db'] = db
                if ':' in host:
                    host, port = host.split(':', 1)
                    kw_args['port'] = int(port)
                kw_args['host'] = host
            else:
                kw_args['db'] = db_host
            if kw_args['db'] and kw_args['db'][0] in ('+', '-'):
                flags['try_transactions'] = kw_args['db'][0]
                kw_args['db'] = kw_args['db'][1:]
            else:
                flags['try_transactions'] = None
            if not kw_args['db']:
                del kw_args['db']
            if items:
                kw_args['user'], items = items[0], items[1:]
            if items:
                kw_args['passwd'], items = items[0], items[1:]
            if items:
                kw_args['unix_socket'], items = items[0], items[1:]
        if timeout:
            kw_args['connect_timeout'] = timeout

        return flags

    def tables(self, rdb=0, _care=('TABLE', 'VIEW')):
        """ Returns list of tables.
        """
        t_list = []
        db_result = self._query('SHOW TABLE STATUS')
        charset = self._kw_args.get('charset', 'UTF-8')
        if charset.startswith('utf8'):
            charset = 'UTF-8'

        for row in db_result.fetch_row(0):
            variables = {}
            for index, key in ((0, 't_name'), (1, 't_engine'), (4, 't_size'),
                               (14, 't_cs')):
                value = row[index]
                if isinstance(value, six.binary_type):
                    value = value.decode(charset)
                variables[key] = value

            description = ('%(t_engine)s, %(t_size)s rows, '
                           'character set/collation %(t_cs)s' % variables)
            t_list.append({'table_name': variables['t_name'],
                           'table_type': 'table',
                           'description': description})
        return t_list

    def columns(self, table_name):
        """ Returns list of column descriptions for ``table_name``.
        """
        c_list = []
        try:
            # Field, Type, Null, Key, Default, Extra
            db_result = self._query('SHOW COLUMNS FROM %s' % table_name)
        except ProgrammingError:
            # Table does not exist. Any other error should propagate.
            LOG.warning('columns query for non-existing table %s' % table_name)
            return ()

        charset = self._kw_args.get('charset', 'UTF-8')
        if charset.startswith('utf8'):
            charset = 'UTF-8'

        for Field, Type, Null, Key, Default, Extra in db_result.fetch_row(0):

            if six.PY3:
                # Force-decoding to make the Browse ZMI tab work across
                # all supported Python versions
                if isinstance(Field, six.binary_type):
                    Field = Field.decode(charset)
                if isinstance(Type, six.binary_type):
                    Type = Type.decode(charset)
                if isinstance(Null, six.binary_type):
                    Null = Null.decode(charset)
                if isinstance(Key, six.binary_type):
                    Key = Key.decode(charset)
                if isinstance(Default, six.binary_type):
                    Default = Default.decode(charset)
                if isinstance(Extra, six.binary_type):
                    Extra = Extra.decode(charset)

            info = {'name': Field,
                    'extra': (Extra,),
                    'nullable': (Null == 'YES') and 1 or 0}

            if Default is not None:
                info['default'] = Default
                field_default = "DEFAULT '%s'" % Default
            else:
                field_default = ''

            if '(' in Type:
                end = Type.rfind(')')
                short_type, size = Type[:end].split('(', 1)
                if short_type not in ('set', 'enum'):
                    if ',' in size:
                        info['scale'], info['precision'] = map(
                            int, size.split(',', 1))
                    else:
                        info['scale'] = int(size)
            else:
                short_type = Type

            if short_type in field_icons:
                info['icon'] = short_type
            else:
                info['icon'] = icon_xlate.get(short_type, 'what')

            info['type'] = short_type
            nul = (Null == 'NO' and 'NOT NULL' or '')
            info['description'] = ' '.join([Type,
                                            field_default,
                                            Extra or '',
                                            key_types.get(Key, Key or ''),
                                            nul])
            if Key:
                info['index'] = True
                info['key'] = Key
            if Key == 'PRI':
                info['primary_key'] = True
                info['unique'] = True
            elif Key == 'UNI':
                info['unique'] = True

            c_list.append(info)

        return c_list

    def variables(self):
        """ Return dictionary of current mysql variable/values.
        """
        # variable_name, value
        variables = self._query('SHOW VARIABLES')
        return dict((name, value) for name, value in variables.fetch_row(0))

    def _query(self, query, force_reconnect=False):
        """
          Send a query to MySQL server.
          It reconnects automaticaly if needed and the following conditions are
          met:
           - It has not just tried to reconnect (ie, this function will not
             attempt to connect twice per call).
           - This conection is not transactionnal and has set no MySQL locks,
             because they are bound to the connection. This check can be
             overridden by passing force_reconnect with True value.
        """
        try:
            self.db.query(query)
        except OperationalError as exc:
            if exc.args[0] in query_syntax_error:
                raise OperationalError(exc.args[0],
                                       '%s: %s' % (exc.args[1], query))

            if not force_reconnect and \
               (self._mysql_lock or self._transactions) or \
               exc.args[0] not in hosed_connection:
                if len(query) > 2000:
                    msg = '%s... (truncated at 2000 chars)' % query[:2000]
                else:
                    msg = query
                LOG.warning('query failed:\n%s' % msg)
                raise

            # Hm. maybe the db is hosed.  Let's restart it.
            if exc.args[0] in hosed_connection:
                msg = '%s Forcing a reconnect.' % hosed_connection[exc.args[0]]
                LOG.error(msg)
            self._forceReconnection()
            self.db.query(query)
        except ProgrammingError as exc:
            if exc.args[0] in hosed_connection:
                self._forceReconnection()
                msg = '%s Forcing a reconnect.' % hosed_connection[exc.args[0]]
                LOG.error(msg)
            else:
                if len(query) > 2000:
                    msg = '%s... (truncated at 2000 chars)' % query[:2000]
                else:
                    msg = query
                LOG.warning('query failed:\n%s' % msg)
            raise

        return self.db.store_result()

    def query(self, sql_string, max_rows=1000):
        """ Execute ``sql_string`` and return at most ``max_rows``.
        """
        self._use_TM and self._register()
        desc = None
        rows = ()

        for qs in filter(None, [q.strip() for q in sql_string.split('\0')]):
            qtype = qs.split(None, 1)[0].upper()
            if qtype == 'SELECT' and max_rows:
                qs = '%s LIMIT %d' % (qs, max_rows)
            db_results = self._query(qs)

            if desc is not None and \
               db_results and \
               db_results.describe() != desc:
                msg = 'Multiple select schema are not allowed.'
                raise ProgrammingError(msg)

            if db_results:
                desc = db_results.describe()
                rows = db_results.fetch_row(max_rows)
            else:
                desc = None

            if qtype == 'CALL':
                # For stored procedures, skip the status result
                self.db.next_result()

        if desc is None:
            return (), ()

        items = []
        for info in desc:
            items.append({'name': info[0],
                          'type': self.defs.get(info[1], 't'),
                          'width': info[2],
                          'null': info[6]})

        return items, rows

    def string_literal(self, sql_str):
        """ Called from zope to quote/escape strings for inclusion
            in a query.
        """
        return self.db.string_literal(sql_str)

    def unicode_literal(self, sql_str):
        """ Similar to string_literal but encodes it first.
        """
        return self.db.unicode_literal(sql_str)

    # Zope 2-phase transaction handling methods

    def _register(self):
        """ Override to replace transaction register() call with join().

            The primary reason for this is to eliminate issues in both the
            super's _register() and transaction register(). The former has a
            bare except that hides useful errors. The latter deals poorly with
            exceptions raised from the join due to state modifications made
            before the join attempt.

            They also both (for our purposes) needlessly add wrapper objects
            around self, resulting in unnecessary overhead.
        """
        if not self._registered:
            try:
                transaction.get().join(self)
            except TransactionFailedError:
                msg = 'database connection failed to join transaction.'
                LOG.error(msg)
            except ValueError:
                msg = 'database connection failed to join transaction.'
                LOG.error(msg, exc_info=True)
                raise
            else:
                self._begin()
                self._registered = True
                self._finalize = False

    def _begin(self, *ignored):
        """ Begin a transaction, if transactions are enabled.

        Also called from _register() upon first query.
        """
        try:
            self._transaction_begun = True
            self.db.ping()
            if self._transactions:
                self._query('BEGIN')
            if self._mysql_lock:
                self._query("SELECT GET_LOCK('%s',0)" % self._mysql_lock)
        except Exception:
            LOG.error('exception during _begin', exc_info=True)
            raise ConflictError

    def _finish(self, *ignored):
        """ Commit a transaction, if transactions are enabled and the
        Zope transaction has committed successfully.
        """
        if not self._transaction_begun:
            return
        self._transaction_begun = False
        try:
            if self._mysql_lock:
                self._query("SELECT RELEASE_LOCK('%s')" % self._mysql_lock)
            if self._transactions:
                self._query('COMMIT')
        except Exception:
            LOG.error('exception during _finish', exc_info=True)
            raise ConflictError

    def _abort(self, *ignored):
        """ Roll back the database transaction if the Zope transaction
        has been aborted, usually due to a transaction error.
        """
        if not self._transaction_begun:
            return
        self._transaction_begun = False
        if self._mysql_lock:
            self._query("SELECT RELEASE_LOCK('%s')" % self._mysql_lock)
        if self._transactions:
            self._query('ROLLBACK')
        else:
            LOG.error('aborting when non-transactional')

    def _mysql_version(self):
        """ Return mysql server version.
        """
        if getattr(self, '_version', None) is None:
            self._version = self.variables().get('version')
        return self._version

    def savepoint(self):
        """ Basic savepoint support.
        """
        if not self._transaction_begun:
            LOG.error('Savepoint used outside of transaction.')
            raise AttributeError

        return _SavePoint(self)


class _SavePoint(object):
    """ Simple savepoint object
    """

    def __init__(self, db_conn):
        self.db_conn = db_conn
        self.ident = str(time.time()).replace('.', 'sp')
        db_conn._query('SAVEPOINT %s' % self.ident)

    def rollback(self):
        self.db_conn._query('ROLLBACK TO %s' % self.ident)



##############################################################################
#
# Copyright (c) 2001 Zope Foundation and Contributors.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
""" Tests for the utils module
"""
import unittest

from .dummy import COLUMNS
from .dummy import TABLE
from .dummy import FakeColumns


class BrowserTestsBase(unittest.TestCase):

    def test_getattr_attributeerror(self):
        browser = self._makeOne()
        with self.assertRaises(AttributeError):
            browser.unknown


class TableBrowserTests(BrowserTestsBase):

    def _getTargetClass(self):
        from Products.ZMySQLDA.utils import TableBrowser
        return TableBrowser

    def _makeOne(self):
        browser = self._getTargetClass()()
        browser._d = TABLE
        browser._c = FakeColumns(TABLE['table_name'])
        return browser

    def test_instantiation(self):
        browser = self._makeOne()
        self.assertEqual(browser.icon, 'what')
        self.assertEqual(browser.description(), '')
        self.assertEqual(browser.check, '')

    def test_getattr(self):
        browser = self._makeOne()
        self.assertEqual(browser.table_name, 'table1')

    def test_tpValues(self):
        browser = self._makeOne()
        cols = browser.tpValues()
        self.assertEqual(cols[0].table_name, 'table1')
        self.assertEqual(cols[0].icon, 'icon1')
        self.assertEqual(cols[1].table_name, 'table1')
        self.assertEqual(cols[1].icon, 'icon2')

    def test_tpId(self):
        browser = self._makeOne()
        self.assertEqual(browser.tpId(), 'table1')

    def test_tpURL(self):
        browser = self._makeOne()
        self.assertEqual(browser.tpURL(), 'Table/table1')

    def test_name(self):
        browser = self._makeOne()
        self.assertEqual(browser.name(), 'table1')

    def test_type(self):
        browser = self._makeOne()
        self.assertEqual(browser.type(), 'type1')


class ColumnBrowserTests(BrowserTestsBase):

    def _getTargetClass(self):
        from Products.ZMySQLDA.utils import ColumnBrowser
        return ColumnBrowser

    def _makeOne(self):
        browser = self._getTargetClass()()
        browser._d = COLUMNS[0]
        browser.table_name = 'table1'
        return browser

    def test_instantiation(self):
        browser = self._makeOne()
        self.assertEqual(browser.icon, 'field')

    def test_getattr(self):
        browser = self._makeOne()
        self.assertEqual(browser.name, 'col1')

    def test_check(self):
        browser = self._makeOne()
        self.assertIn('table1.col1', browser.check())

    def test_tpId(self):
        browser = self._makeOne()
        self.assertEqual(browser.tpId(), 'col1')

    def test_tpURL(self):
        browser = self._makeOne()
        self.assertEqual(browser.tpURL(), 'Column/col1')

    def test_description(self):
        browser = self._makeOne()
        self.assertEqual(browser.description(), 'desc1')


class valuesTests(unittest.TestCase):

    def _makeOne(self):
        from Products.ZMySQLDA.utils import values
        return values()

    def test_len(self):
        vals = self._makeOne()
        self.assertEqual(vals.len(), 1)

    def test_getitem_firstaccess(self):
        # On first access, the ``_d`` attribute is created
        vals = self._makeOne()

        def _callback():
            return ['foo', 'bar', 'baz']

        with self.assertRaises(AttributeError):
            vals._d

        vals._f = _callback
        self.assertEqual(vals[0], 'foo')


def test_suite():
    return unittest.TestSuite((unittest.makeSuite(TableBrowserTests),
                               unittest.makeSuite(ColumnBrowserTests),
                               unittest.makeSuite(valuesTests)))
